import re
import sys
# from time import sleep

from chart import Chart
from symbols import Symbol
from log import log, set_log_level, DEBUG
from rules import Rules
from vertical import XmlTag


class ChartParser:
    def __init__(self, rules=None, rule_files=['default'], argv=[]):
        self.options = dict(
            debug=False,  # print under the hood information
            chart=False,
            edges=False,  # show each edge as it is handled and combined
            # original only, no dependencies yet (--dependency/-d)
            vertical=False,  # TODO: accept file names?
            phrasal_trees=False,  # constituency_trees
            dot=False,  # trees in the DOT format
            vertical_phrases=False,  # TODO: accept file names?
            interactive=False,  # stop after each processed sentence
        )

        argv_rule_files = []
        expect_rules = False
        for arg in argv[1:]:
            if arg == '--rules':
                expect_rules = True
            elif expect_rules:
                argv_rule_files.append(arg)
                expect_rules = False
            elif arg == '--debug':
                self.options['debug'] = True
            elif arg == '--edges':
                self.options['edges'] = True
            elif arg == '--vertical':
                self.options['vertical'] = True
            elif arg == '--chart':
                self.options['chart'] = True
            elif arg == '--trees':
                self.options['phrasal_trees'] = True
            elif arg == '--dot':
                self.options['dot'] = True
            elif arg.startswith('--vertical-phrases'):
                self.options['vertical_phrases'] = arg[19:] or True
                self.options['vertical'] = True
            # elif arg == '--interactive':
            #     self.options['interactive'] = True
            else:
                log.warning('Unknown parameter: ' + arg)

        if self.options['debug']:
            set_log_level(DEBUG)

        if argv_rule_files or rule_files:
            self.rules = Rules(files=argv_rule_files or rule_files)
        elif rules:
            self.rules = Rules(rules)
        self.rules.print()

    def prepare(self, tokens, root):
        self.tokens = tokens

        for seq, token in enumerate(tokens):
            token.parse_tag()
            token.seq = seq
            log.info(Symbol(token))

        self.root = root
        # když mezi nima není sloveso nebo „lze“ k6, spustit parse(root='NP')
        # teď to není potřeba, protože CLAUSE predikuje i NP
        # ale bude se to hodit na vernacular a optional_punctuation
        if False:
            log.info('No verb found, parsing as NP instead')

    def parse(self, tokens, root='CLAUSE', opening_tag=None, closing_tag=None):
        self.prepare(tokens, root)

        length = len(self.tokens)
        self.known_edges = set()
        self.prepare_agenda()
        self.chart = Chart(length)

        while self.agenda:
            agenda_edge = self.agenda.popleft()
            for new_edge in self.propose_edges(agenda_edge):
                self.agenda.append(new_edge)
            self.chart.append(agenda_edge)

        if self.options['chart']:
            self.chart.print()
        self.successful_edges = [
            edge for edge in self.chart.closed_edges[0] if edge.end == length
            ]
        if not self.successful_edges:
            log.error('\nNo edge covering the whole input!\n')
            # TODO: vytisknout nejdelší hrany (nejzazší edge.end)
        if opening_tag is not None:
            opening_tag['passed'] = 'yes' if self.successful_edges else 'no'
        for edge in self.successful_edges:
            log.info('\nPassed!      %s\n', edge)
            if self.options['phrasal_trees']:
                self.print_graph(edge)
                log.info('\n')
            if self.options['dot']:
                self.print_dot(edge)
            if self.options['vertical_phrases']:
                # CLAUSE, ATTR_CLAUSE?, DEP_CLAUSE
                self.print_vertical_phrases(edge, opening_tag, closing_tag,
                                            self.options['vertical_phrases'])
        if not self.successful_edges or (self.options['vertical'] and
                not self.options['vertical_phrases']):
            self.print_plain_vertical(opening_tag, closing_tag)
        # if self.interactive:
        #     input()

    def parse_from_vertical(self, lines=sys.stdin, root='CLAUSE'):
        tokens = []
        opening_tag = None
        phrase = None
        seq = 1

        for line in lines:
            line = line.strip()
            if line.startswith('<'):
                tag = XmlTag.parse(line)
                if tag.tag in ('s', 'head'):
                    closing_tag = tag if tag.closing else None
                    if closing_tag is not None or tokens:
                        self.parse(tokens, opening_tag=opening_tag,
                                   closing_tag=closing_tag)
                        seq += 1
                        tokens = []
                        # sleep(0.1)
                    opening_tag = tag if tag.opening else None
                    if tag.opening:
                        opening_tag['seq'] = seq
                    continue

                if tag.tag == 'g':  # and tag.empty
                    if tokens:
                        # the next token is “glued” to this one
                        tokens[-1].trailing_whitespace = False
                elif tag.tag == 'phr':  # MWE
                    if tag.opening:
                        phrase = Symbol(word=[], lemma=tag.pop('l', None),
                                        tag=tag.pop('t', None), mwe=True)
                    elif phrase:
                        phrase['word'] = ' '.join(phrase['word'])
                        tokens.append(phrase)
                        phrase = None
                    else:
                        log.warning('unexpected </phr>!')
                elif tag.tag == 'doc':
                    if tokens:
                        self.parse(tokens, opening_tag=opening_tag)
                        seq += 1
                        tokens = []
                    if self.options['vertical']:
                        print(tag)
                elif tag.tag == 'p':
                    if tokens:
                        surrogate_opening_tag = (
                            tag if opening_tag is None and tag.opening
                            else opening_tag)
                        surrogate_opening_tag['seq'] = seq
                        self.parse(tokens, opening_tag=surrogate_opening_tag)
                        seq += 1
                        tokens = []
                    elif tag.closing and self.options['vertical']:
                        print(tag)
                elif tag.tag == 'table':
                    # přeskakovat někdy celý obsah?
                    # v DESAMu source.130620.294.utf8 je už jen jednou </table>
                    pass  # table = True
                elif tag.tag == 'sign':
                    pass  # zajímavé: Josef Juřeník, přednosta stanice
                    # neboli přístavek
            elif line:
                if phrase is not None:
                    phrase['word'].append(line)
                else:
                    try:
                        word, lemma, tag, *extra = line.split()
                    except ValueError:
                        word = lemma = line
                        extra = []
                        if line in ',.():?':
                            tag = 'kI'  # cesta do pekel :-D
                        # TODO: vylepšit heuristiku
                        elif line[0] in '0123456789':
                            log.warning('line with a number, correct? ' + line)
                            tag = 'k4'
                        else:  # 31. března
                            log.warning('unexpected line ' + line)
                            tag = 'k?'
                    token = Symbol(word=word, lemma=lemma, tag=tag,
                                   extra=extra, line=line)
                    tokens.append(token)
        if tokens:
            self.parse(tokens, root=root)  # TODO: opening_tag?

    def print_graph(self, edge, level=0):
        log.info('{0}{1}'.format('   '*level, edge))
        level += 1
        for child in edge:
            if child.edge is True:
                if len(edge) > 1:
                    log.info('{0}{1}'.format('   '*level, child))
            elif child.edge and child.length:
                self.print_graph(child.edge, level)

    def print_dot(self, edge, parent_indexes='0'):
        if parent_indexes == '0':
            print('graph {')
        print('    {} [label="{}"];'.format(parent_indexes, edge.left))
        for index, child in enumerate(edge):
            if child.edge is True:
                print('    {}{} [label="{}"];'.format(
                    parent_indexes, index, child))
            elif child.edge and child.length:
                self.print_dot(child.edge, parent_indexes + str(index))
            else:
                continue
            print('    {0} -- {0}{1};'.format(parent_indexes, index))
        if parent_indexes == '0':
            print('}')

    def print_plain_vertical(self, opening_tag, closing_tag):
        if opening_tag is not None:
            print(opening_tag)
        for token in self.tokens:
            print(token.vertical())
        if closing_tag is not None:
            print(closing_tag)

    def print_vertical_phrases(self, edge, opening_tag=None, closing_tag=None,
                               phrases=True):
        if opening_tag is not None:
            print(opening_tag)
        phrase = edge.left['phrase'].lower()
        if phrases is True or re.match(phrases, phrase):
            phrase = XmlTag(tag=phrase, opening=True)
            phrase.update(
                (attr, value) for (attr, value) in edge.left.items()
                if attr not in('phrase',) and value)
            print(phrase)
            phrase.opening = False
            phrase.clear()
        else:
            phrase = None
        for child in edge:
            if child.edge is True:
                if len(edge) >= 1:
                    print(child.vertical())
            elif child.edge and child.length:
                self.print_vertical_phrases(child.edge, phrases=phrases)
        if phrase is not None:
            print(phrase)
        if closing_tag is not None:
            print(closing_tag)
