from itertools import chain
import re

from log import log
from shlex import shlex  # shlex je teď můj (a mám svoje escape místo quote)
import vertical


ORDER = 'kegncpamdxytzw~'
ALL_META = ('line', 'edge', 'optional', 'begin', 'length', 'head', 'dep',
            'mwe', 'trailing_whitespace', 'extra', 'seq')


class Symbol(dict):  # OrderedDict to zpomaluje snad dvakrát…
    def __init__(self, symbol=None, line=None, edge=None, optional=None,
                 begin=None, length=None, head=None, dep=None, mwe=None,
                 trailing_whitespace=True, extra=None, seq=None, **kwargs):
        if symbol is not None:
            if isinstance(symbol, str):
                symbols = read_rule(symbol)
                try:
                    symbol = next(symbols)
                except StopIteration:
                    raise ValueError('No symbol passed: "' + symbol + '"')
                # '•' neprojde
                extra = list(symbols)
                if extra:
                    # TODO: taky ValueError?
                    log.warning('Extra symbols passed, ignoring them: %s',
                                ' '.join(str(t) for t in extra))
            super().__init__(symbol)
            self.line = symbol.line
            self.extra = symbol.extra
            self.seq = symbol.seq
            self.edge = symbol.edge
            self.optional = symbol.optional
            self.begin = symbol.begin
            self.length = symbol.length
            self.head = symbol.head
            self.dep = symbol.dep
            self.mwe = symbol.mwe
            self.trailing_whitespace = symbol.trailing_whitespace
        else:
            super().__init__(**kwargs)
            self.line = line
            self.extra = extra  # další sloupce z vertikálu (třeba závislost)
            self.seq = seq  # pořadové číslo
            self.edge = edge
            self.optional = optional
            self.begin = begin
            self.length = length
            self.head = head
            self.dep = dep
            self.mwe = mwe
            self.trailing_whitespace = trailing_whitespace
        # TODO: ještě nevím, jestli nepovinné/prázdné fráze vždycky skrývat
        # self.hidden = False

    @property
    def end(self):
        return self.begin + self.length

    def __eq__(self, other):
        if isinstance(other, str):
            return False
        elif self.begin != other.begin or self.length != other.length:
            return False
        else:
            return super().__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __format__(self, format_spec=''):  # 's' for (equal) span, 't' for tag
        if self.mwe:
            phr = vertical.XmlTag(tag='phr', opening=True)
            phr.update(t=self.get('tag', ''), l=self.get('lemma', ''))
            return (str(phr) + '\n' + self.get('word', '').replace(' ', '\n') +
                    '\n</phr>')

        if not format_spec:
            equal_span = False
            print_tag = False
        else:
            equal_span = 's' in format_spec
            print_tag = 't' in format_spec

        phrase_head = self.get('phrase')
        phrase_head = [phrase_head] if phrase_head else []
        is_head = ['head'] if self.head else []
        dependency = ['dep=' + escape(self.dep)] if self.dep else []
        word = self.get('word')
        wlt = ['word=' + escape(word)] if word else []
        lemma = self.get('lemma')
        tag = self.get('tag')
        if lemma:
            wlt.append('lemma=' + escape(lemma))
        if print_tag and tag:
            wlt.append('tag=' + escape(tag))
        short_attrs = []
        for attr in ORDER:
            # TODO: fakt je asi lepší shodu ukládat jako True
            value = self.get(attr, -123)
            if value is None:
                short_attrs.append(attr)
            elif value != -123:
                short_attrs.append(attr + '=' + escape(value))
        span = ['%s–%s' % (self.begin, self.end)] if (
            self.length is not None and (self.length or equal_span)) else []
        optional = ')?' if self.optional else ')'
        return ('(' + ' '.join(attr_value for attr_value in chain(
            phrase_head, is_head, wlt, short_attrs, dependency, span)) +
            optional)

    def __str__(self):
        return format(self, 's')

    def __repr__(self):
        return 'Symbol(' + escape(self) + ')'

    def update(self, symbol=None, **kwargs):
        if symbol is not None:
            super().update(symbol)
        if isinstance(symbol, Symbol):
            for meta in ALL_META:
                self.__dict__[meta] = symbol.__dict__[meta]
        for attr in dict(kwargs):
            if attr in ALL_META:
                self.__dict__[attr] = kwargs
                kwargs.pop(attr, None)

    def parse_tag(self):
        tag = self.get('tag')
        if tag:
            self.update(vertical.extract_attributes_from_tag(tag))

    def vertical(self):
        glue = '' if self.trailing_whitespace else '\n<g/>'
        return '{}\t{}\t{}\t{}{}'.format(
            self.get('word', ''), self.get('lemma', ''), self.get('tag', ''),
            self.seq, glue)


class Symbols(list):
    """Abstrakce pro pravou stranu pravidla/hrany (anebo její část)"""
    def __init__(self, iterable=None, position=None):
        self.position = position
        self.stringified = None
        if isinstance(iterable, str):
            self.extend(read_rule(iterable))
        elif iterable is not None:
            super().__init__((Symbol(s) for s in iterable))
        else:
            super().__init__()

    def __add__(self, value):
        if isinstance(value, Symbol):
            self.append(value)
        elif isinstance(value, str):
            self.extend(Symbols(value))
        else:
            return super().__add__(value)

    def __format__(self, format_spec=''):  # full (excluding parsed tags)
        if not self:
            return 'ε' if self.position is None else 'ε •'

        # if self.position is None or format_spec is None:
        symbols = [format(symbol, format_spec) for symbol in self]
        if self.position is not None:
            symbols.insert(self.position, '•')
        return ' '.join(symbols)
        # else:
        #     symbols = [(str(symbol) if index == self.position or
        #                 symbol.length else '') for index, symbol
        #                in enumerate(self)]
        #     symbols.insert(self.position, '•')
        #     return ' '.join(symbol for symbol in symbols if symbol)

    def __str__(self):
        if self.stringified is None:
            self.stringified = format(self)
        return self.stringified

    def __repr__(self):
        return 'Symbols(' + escape(self) + ')'


def read_rule(rule):
    """
    Return a list of symbols, i.e. (non-)terminals from a string
    (which has already been split to the left/right side)
    """
    if rule.strip() == 'ε':
        return

    symbol = None  # not None ⇔ inside the symbol
    head = False  # expecting the phrase head (or nothing)
    attr = ''  # a = plus an attribute value may follow
    equal_sign = False  # we are expecting a value for an attribute after =
    closed = False  # the symbol has already ended with )

    for token in shlex(rule, posix=True):
        if token == '(' or (token == '•' and (symbol is None or closed)):
            if symbol is not None:
                yield symbol
                if not closed:  # TODO: allow ambiguous (a=( b=1) ?
                    raise ValueError('Unexpected "("')

            if token == '•':
                yield '•'
                symbol = None
                continue

            symbol = Symbol()
            head = closed = False
        elif token == '?':
            symbol.optional = True
        elif token == ')':
            if equal_sign:
                raise ValueError('Unterminated attribute ' + attr)
            elif attr == 'head':
                symbol.head = True
            elif attr:
                symbol[attr] = None  # není lepší pro shodové atributy True?

            attr = ''
            equal_sign = False
            closed = True
        elif closed:
            raise ValueError('Not expected outside a symbol: ' + token)
        elif not attr:
            if token == '=':
                raise ValueError('Unexpected "="')
            elif not head and token[0].isupper():
                head = symbol['phrase'] = token
                continue
            attr = token
        elif token == '=':
            equal_sign = True
        elif equal_sign:
            if attr == 'dep':
                symbol.dep = token
            elif attr == 'head':
                symbol.head = token  # stačil by bool
            else:
                symbol[attr] = token
            attr = ''
            equal_sign = False
        else:
            if attr == 'head':
                symbol.head = True
            else:
                symbol[attr] = None
            attr = token

    if symbol is not None:
        yield symbol
        if not closed:
            raise ValueError('Rule not terminated by ")"')


need_escaping = re.compile(r'[()=? \t\n]').search  # podle shlex._find_unsafe


def escape(s):  # začátek podle shlex.quote
    if not isinstance(s, str):
        s = str(s)
    if not s:
        return "''"
    elif need_escaping(s) is None:
        return s
    elif "'" not in s:
        return "'" + s + "'"
    elif '"' not in s:
        return '"' + s + '"'
    else:
        raise ValueError("Can't escape, the string includes both quotes: " + s)
